MODULE MakeSquare EXPORTS Main;

IMPORT Oct, Color, Triang, Math, LR3, Random, Wr, Stdio, ParseParams, 
       Process, Thread;
FROM Oct IMPORT Lnext, Sym, Rot;
FROM Triang IMPORT Topology, Read, Write, Coords, OrgV, LeftF;
FROM Stdio IMPORT stderr;

TYPE
  Shape = {Flat, Bent};
  Options = RECORD
      inFile: TEXT;  (* Must have been created with MakeRawCube *)
      outFile: TEXT;
      gridOrder: CARDINAL;
      shape: Shape;
      random: BOOLEAN;
      fixBorder: BOOLEAN;
    END;

PROCEDURE Main() =
  BEGIN
    WITH 
      o = GetOptions(),
      tc = Read(o.inFile),
      top = tc.top,
      c = tc.c^
    DO
      SetAttributes(top, c, o.gridOrder, o.shape, o.random, o.fixBorder);
      Write(o.outFile, top, c)
    END;
  END Main;

TYPE
  VertexLabel = ARRAY [0..2] OF INTEGER;

PROCEDURE CoordsToLabel(READONLY vc: LR3.T): VertexLabel =
  BEGIN
    WITH
      ix = ROUND(vc[0]),
      iy = ROUND(vc[1]),
      iz = ROUND(vc[2])
    DO
      RETURN VertexLabel{ix, iy, iz}
    END;
  END CoordsToLabel;
  
PROCEDURE ThrowPoint(coins: Random.T; avg, rad: LR3.T): LR3.T =
  (* Generates a random point in the box with center "avg" and
  half-zise "rad". *)
  VAR res: LR3.T;
  BEGIN
    FOR i := 0 TO 2 DO 
      res[i] := avg[i] + rad[i] * coins.longreal(-1.0d0, +1.0d0)
    END;
    RETURN res
  END ThrowPoint;

PROCEDURE SetAttributes(
    READONLY top: Topology;
    VAR c: Coords;
    gridOrder: CARDINAL;
    shape: Shape; 
    random: BOOLEAN;
    fixBorder: BOOLEAN;
  ) =
  BEGIN
    WITH 
      MaxLabel = 2*gridOrder,
      Radius = FLOAT(gridOrder, LONGREAL),
      Size = FLOAT(2*gridOrder, LONGREAL),
      Angle = FLOAT(Math.Pi, LONGREAL)/Size,
      coins = NEW(Random.Default).init(TRUE)
    DO 
      (* Set vertex properties *)
      FOR i := 0 TO top.NV-1 DO 
        WITH
          v = top.vertex[i],
          vc = c[v.num],
          vLabel = CoordsToLabel(vc)
        DO
          <* ASSERT 
            MIN(MIN(vLabel[0],vLabel[1]),vLabel[2]) = 0 OR 
            MAX(MAX(vLabel[0],vLabel[1]),vLabel[2]) = MaxLabel 
          *>
          IF vLabel[2] = MaxLabel THEN
            WITH 
              distBorder = MIN(
                MIN(vLabel[0], MaxLabel-vLabel[0]), 
                MIN(vLabel[1], MaxLabel-vLabel[1])
              )
            DO
              v.fixed := distBorder = 0 OR (fixBorder AND distBorder <= 2);
              v.exists := TRUE;
              v.radius := 0.02;
              v.color := Color.T{0.60, 0.80, 1.00};
            END;
          ELSE
            v.fixed := FALSE;
            v.exists := FALSE
          END
        END;
      END;
      
      (* Set face properties *)
      FOR i := 0 TO top.NF-1 DO
        WITH f = top.face[i] DO 
          VAR 
            exists: BOOLEAN := TRUE; 
            p: Triang.Arc := Rot(top.side[i]);
          BEGIN
            REPEAT
              exists := exists AND OrgV(p).exists;
              p := Lnext(p)
            UNTIL p = Rot(top.side[i]);
            f.exists := exists;
            IF exists THEN 
              f.patch := 1;
              f.color := Color.T{1.00, 0.95, 0.70};
              f.transp := Color.T{0.80, 0.80, 0.80};
            END;
          END
        END;
      END;
      
      (* Set edge properties *)
      FOR i := 0 TO top.NE-1 DO 
        WITH a = top.edge[i], e = NARROW(a.edge, Triang.Edge) DO 
          e.exists := OrgV(a).exists AND OrgV(Sym(a)).exists;
          e.spring := LeftF(a).exists AND LeftF(Sym(a)).exists;
          e.color := Color.T{0.00, 0.10, 0.70};
          e.radius := 0.01;
        END
      END;
      
      (* Set vertex coordinates *)
      FOR i := 0 TO top.NV-1 DO 
        WITH
          v = top.vertex[i],
          vc = c[v.num],
          vLabel = CoordsToLabel(vc)
        DO
          VAR cNew: LR3.T;
          BEGIN
            IF vLabel[2] = MaxLabel AND shape = Shape.Bent THEN
              cNew := LR3.T{
                vc[0], 
                Radius - Radius * Math.cos(Angle*FLOAT(vc[1], LONGREAL)),
                Size + Radius * Math.sin(Angle*FLOAT(vc[1], LONGREAL))
              };
            ELSE 
              cNew := vc;
            END;
            
            cNew := LR3.Add(
              LR3.Scale(1.0d0/FLOAT(gridOrder, LONGREAL), cNew), 
              LR3.T{-1.0d0, -1.0d0, -2.0d0}
            );                  
           
            IF vLabel[2] = MaxLabel AND (NOT v.fixed) AND random THEN
              cNew := LR3.Add(
                cNew,
                ThrowPoint(
                  coins, 
                  avg := LR3.T{0.0d0, 0.0d0, 0.0d0},
                  rad := LR3.T{1.5d0, 1.5d0, 1.5d0}
                )
              )
            END;
            vc := cNew
          END
        END
      END
    END
  END SetAttributes;

PROCEDURE GetOptions (): Options =
  <* FATAL Thread.Alerted, Wr.Failure *>
  VAR o: Options;
  BEGIN
    WITH pp = NEW(ParseParams.T).init(stderr) DO
      TRY

        pp.getKeyword("-inFile");                               
        o.inFile := pp.getNext();                         

        pp.getKeyword("-outFile");                               
        o.outFile := pp.getNext();                         

        pp.getKeyword("-gridOrder");                               
        o.gridOrder := pp.getNextInt(1, 20);                         

        IF pp.keywordPresent("-shape") THEN
          IF pp.testNext("flat") THEN
            o.shape := Shape.Flat
          ELSIF pp.testNext("bent") THEN
            o.shape := Shape.Bent
          ELSE
            pp.error("Invalid shape \"" & pp.getNext() & "\"\n");
          END
        ELSE 
          o.shape := Shape.Flat
        END;
        o.random := pp.keywordPresent("-random"); 
        o.fixBorder := pp.keywordPresent("-fixBorder");

        pp.finish();                                       
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: MakeSquare\\\n");
          Wr.PutText(stderr, "  -inFile <name> -outFile <name>\\\n");
          Wr.PutText(stderr, "  -gridOrder <num>\\\n");        
          Wr.PutText(stderr, "  [ -shape flat | bent ]\\\n");
          Wr.PutText(stderr, "  [ -random ]\\\n");
          Wr.PutText(stderr, "  [ -fixBorder ]\n");        
          Process.Exit (1);                                              
      END
    END;
    RETURN o
  END GetOptions;

BEGIN 
  Main();
END MakeSquare.
