MODULE RefineTriang EXPORTS Main;

(*
  Applies smoothing heuristics to a given ".top" file.
  
  Created 1994 by Rober M. Rosi and J. Stolfi.
*)

IMPORT Thread, Wr, Process;
IMPORT LR3, Triang, Oct, ParseParams;
FROM Triang IMPORT 
  Coords, Vertex, OrgV, MakeEdge, MakeVertex, MakeFace,
  SetOrg, SetLeft;
FROM Oct IMPORT
  Arc, Rot, Tor, Flip, Sym, Onext, Oprev, Splice, 
  Lnext, FlipBit, SenseBit, Degree;

FROM Stdio IMPORT stderr;

TYPE
  Options = RECORD              (* Input file name (minus ".top" extension) *)        
      inFile: TEXT;             (* Output file name prefix *)                         
      outFile: TEXT;            (* Number of configurations to generate *) 
      fixOriginal: BOOLEAN;     (* TRUE to fix original vertices *)
    END;
    
PROCEDURE DoIt() =

  CONST 
    ThinEdgeRadius = 0.01;
    
  VAR
    NNE: CARDINAL := 0;

  BEGIN
    WITH 
      o = GetOptions(), 
      tc = Triang.Read(o.inFile),
      top = tc.top,
      c = tc.c^,
      half = NEW(REF ARRAY OF Arc, 2*top.NE)^,
      vnew = NEW(REF ARRAY OF Vertex, top.NV)^
    DO

      PROCEDURE Half(a: Arc): Arc =
        BEGIN
          WITH
            ia = a.edge.num,
            sa = SenseBit(a),
            fa = FlipBit(a),
            h = half[2*ia + sa]
          DO
            IF fa = 0 THEN RETURN h ELSE RETURN Flip(h) END
          END
        END Half;

      BEGIN
        (* Copy original vertices, save correspondence in "vnew": *)
        FOR iu := 0 TO top.NV-1 DO 
          WITH
            u = top.vertex[iu],
            v = MakeVertex()
          DO
            v.num := u.num;
            v.exists := u.exists;
            v.fixed := u.fixed OR o.fixOriginal;
            v.color := u.color;
            v.radius := u.radius;
            vnew[iu] := v
          END
        END;

        (* 
          Create two new edges for each original edge "e".
          The new edge corresponding to the origin half of
          "e", with same direction and orientation, will be
          | Half(e) = Flip^f(half[2*e.edge.num + s])
          where f = FlipBit(e), s = SenseBit(e)
        *)
        FOR ie := 0 TO top.NE-1 DO 
          WITH
            a = top.edge[ie],
            sa = SenseBit(a),
            fa = FlipBit(a),
            ho = half[2*ie + sa],
            hd = half[2*ie + 1 - sa],
            e = NARROW(a.edge, Triang.Edge),
            v = MakeVertex()
          DO
            <* ASSERT e.num = ie *>

            v.num := top.NV + ie;
            v.exists := e.exists;
            v.fixed := FALSE;
            v.radius := e.radius;
            v.color := e.color;

            ho := MakeEdge();
            WITH hoe = NARROW(ho.edge, Triang.Edge) DO
              hoe.num := NNE; INC(NNE);
              hoe.exists := e.exists;
              hoe.spring := e.spring;
              hoe.radius := e.radius;
              hoe.color := e.color;
            END;
            IF fa = 1 THEN ho := Flip(ho) END;

            hd := MakeEdge();
            WITH hde = NARROW(hd.edge, Triang.Edge) DO
              hde.num := NNE; INC(NNE);
              hde.exists := e.exists;
              hde.spring := e.spring;
              hde.radius := e.radius;
              hde.color := e.color;
            END;
            IF fa = 1 THEN hd := Flip(hd) END;

            Splice(Sym(ho), Sym(hd));

            SetOrg(Sym(ho), v);
            SetOrg(Sym(hd), v);
            SetOrg(ho, vnew[OrgV(a).num]);
            SetOrg(hd, vnew[OrgV(Sym(a)).num]);
          END;
        END;

        (* Connect the half-edges as in the original triangulation: *)
        FOR ie := 0 TO top.NE-1 DO 
          WITH
            a = top.edge[ie],
            b = Onext(a),
            ha = Half(a),
            hb = Half(b)
          DO
            IF b # a AND Onext(ha) # hb THEN
              Splice(ha, Oprev(hb))
            END
          END;
          WITH
            a = Sym(top.edge[ie]),
            b = Onext(a),
            ha = Half(a), 
            hb = Half(b)
          DO
            IF b # a AND Onext(ha) # hb THEN
              Splice(ha, Oprev(hb))
            END
          END
        END;

        FOR jf := 0 TO top.NF-1 DO 
          WITH 
            f = top.face[jf], 
            nf = MakeFace(),
            df = Degree(top.side[jf])
          DO
            nf.exists := f.exists;
            nf.color := f.color;
            nf.transp := f.transp;
            nf.patch := f.patch;

            VAR a: Arc := Rot(top.side[jf]); a0: Arc := a;

            BEGIN
              REPEAT
                WITH
                  h = Half(a),
                  m = Half(Onext(a)),
                  s = MakeEdge(),
                  se = NARROW(s.edge, Triang.Edge),
                  g = MakeFace()
                DO

                  se.num := NNE; INC(NNE);
                  se.exists := f.exists;
                  se.color := f.color;
                  se.radius := ThinEdgeRadius;

                  Splice(s, Sym(m));
                  Splice(Sym(s), Oprev(Sym(h)));

                  SetOrg(s, OrgV(Sym(m)));
                  SetOrg(Sym(s), OrgV(Sym(h)));

                  <* ASSERT Degree(Rot(s)) = 3 *>

                  g.exists := f.exists;
                  g.color := f.color;
                  g.transp := f.transp;
                  g.patch := f.patch;

                  SetLeft(Sym(m), g);
                  SetLeft(h, g);
                  SetLeft(Sym(s), g);
                  SetLeft(s, nf);

                END;
                a := Lnext(a)
              UNTIL a = a0;
              WITH
                s = Sym(Oprev(Sym(Half(a0))))
              DO
                <* ASSERT Degree(Tor(s)) = df *>
              END;
            END;
          END
        END;
      
        EVAL Oct.NumberEdges(half);
        (* Number faces: *)
        EVAL Triang.NumberVertices(Rot(Half(top.edge[0])));
        WITH
          ntop = Triang.MakeTopology(Half(top.edge[0])),
          nc = NEW(REF Coords, ntop.NV)^,
          comments = "refined from:\n" & tc.comments
        DO
          <* ASSERT ntop.NV = top.NV + top.NE *>
          <* ASSERT ntop.NF = top.NF + 2*top.NE *>
          <* ASSERT ntop.NE = 4*top.NE *>
          FOR ie := 0 TO top.NE - 1 DO 
            WITH 
              e = top.edge[ie],
              oiu = OrgV(e).num,
              oiv = OrgV(Sym(e)).num,
              niu = OrgV(Half(e)).num, 
              niv = OrgV(Half(Sym(e))).num,
              nix = OrgV(Sym(Half(e))).num,
              niy = OrgV(Sym(Half(Sym(e)))).num
            DO 
              <* ASSERT nix = niy *>
              nc[niu] := c[oiu];
              nc[niv] := c[oiv];
              nc[nix] := LR3.Scale(0.5d0, LR3.Add(c[oiu], c[oiv]))
            END
          END;
          Triang.Write(o.outFile, ntop, nc, comments)
        END
      END;
    END
  END DoIt;

PROCEDURE GetOptions (): Options =
  <* FATAL Thread.Alerted, Wr.Failure *>
  VAR o: Options;
  BEGIN
    WITH pp = NEW(ParseParams.T).init(stderr) DO                         
      TRY
        pp.getKeyword("-inFile");                               
        o.inFile := pp.getNext();                         

        pp.getKeyword("-outFile");                               
        o.outFile := pp.getNext();
        
        o.fixOriginal := pp.keywordPresent("-fixOriginal");

        pp.finish();                                       
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: RefineTriang \\\n");
          Wr.PutText(stderr, "  -inFile <name>  -outFile <name> \\\n");
          Wr.PutText(stderr, "  [ -fixOriginal ] \n");
          Process.Exit (1);                                              
      END
    END;
    RETURN o
  END GetOptions; 
  
BEGIN
  DoIt()
END RefineTriang. 
