MODULE CornerEnergy;
IMPORT Math, LR3, LR3Extras, Triang; 
FROM Oct IMPORT Onext, Sym;
FROM Triang IMPORT Arc, OrgV, LeftF, Topology;
FROM Energy IMPORT Coords, Gradient;

REVEAL
  T = Public BRANDED OBJECT
      top: Topology;
      (* Defined by "defTop" *)
      cosi, sini: REF ARRAY OF LONGREAL;  (* Cos and sin of ideal angle at vertex *)
      K: LONGREAL;                        (* Energy normalization factor *)
      (* Defined by "defVar: *)
      vVar: REF ARRAY OF BOOLEAN;         (* Tells which vertices are variable *)
      termVar: REF ARRAY OF BOOLEAN;      (* Tells which terms must be computed *)
    OVERRIDES
      init := Init;
      defTop := DefTop;
      defVar := DefVar;
      eval := Eval;
      name := Name;
    END;
    
PROCEDURE Init(erg: T): T =
  BEGIN
    RETURN erg
  END Init;

PROCEDURE DefTop(erg: T; READONLY top: Topology) =
  BEGIN
    erg.top := top;
    erg.cosi := NEW(REF ARRAY OF LONGREAL, top.NV);
    erg.sini := NEW(REF ARRAY OF LONGREAL, top.NV);
    erg.K := 1.0d0/FLOAT(16*top.NV, LONGREAL);
    DefineCosISinI(top, erg.cosi^, erg.sini^);
    erg.vVar := NEW(REF ARRAY OF BOOLEAN, top.NV);
    erg.termVar := NEW(REF ARRAY OF BOOLEAN, top.NV);
    (* Just in case the client forgets to call "defVar": *)
    FOR i := 0 TO top.NV-1 DO 
      erg.vVar[i] := FALSE;
      erg.termVar[i] := FALSE
    END;
  END DefTop;
  
PROCEDURE DefineCosISinI(READONLY top: Topology; VAR cosi, sini: ARRAY OF LONGREAL) =
  (*
    Define "cosi[v]" and "sini[v]" as the cosine and sine of the 
    ideal projected angle between adjacent edges at vertex "v".
   
    A vertex matters only if it exists, has degree 
    three or more, and all its incident edges and faces (and hence 
    its neighbors) exist.  Vertices that don't matter have their 
    "sini" and "cosi" set to zero.
  *)
  BEGIN
    WITH
      NV = top.NV,
      out = top.out^,
      vertex = top.vertex^
    DO
      <* ASSERT NUMBER(cosi) = NV *>
      <* ASSERT NUMBER(sini) = NV *>
      FOR v := 0 TO NV-1 DO 
        VAR complete: BOOLEAN := TRUE; d: CARDINAL := 0;
        BEGIN
          IF vertex[v].exists THEN
            WITH ov = out[v] DO
              VAR a: Arc := ov;
              BEGIN 
                REPEAT
                  WITH 
                    e = NARROW(a.edge, Triang.Edge),
                    u = OrgV(Sym(a)), f = LeftF(a)
                  DO
                    complete := complete AND f.exists;
                    IF f.exists THEN <* ASSERT e.exists *> END;
                    IF e.exists THEN <* ASSERT u.exists *> END;
                  END;
                  INC(d); a := Onext(a)
                UNTIL a = ov;
              END
            END
          ELSE
            complete := FALSE; d := 0
          END;
          IF complete AND d >= 3 THEN
            WITH ia = 2.0d0*FLOAT(Math.Pi, LONGREAL)/FLOAT(d, LONGREAL) DO
              cosi[v] := Math.cos(ia);
              sini[v] := Math.sin(ia) 
            END
          ELSE
            cosi[v] := 0.0d0; sini[v] := 0.0d0;
          END;
        END;
      END;    
    END
  END DefineCosISinI;

PROCEDURE DefVar(erg: T; READONLY variable: ARRAY OF BOOLEAN) =
  BEGIN
    (*
      Decide which terms are variable, and compute 
      effective degrees "erg.deg[v]".

      There is a term for each vertex that matters (exists,
      has degree 3 or more, and and has only existing edges 
      and edges incident to it).
      
      The term of "v" is variable if "v" is variable
      or has some neighbor that is variable.
    *)
    WITH
      NV = erg.top.NV,
      out = erg.top.out^,
      vertex = erg.top.vertex^,
      vVar = erg.vVar^,
      termVar = erg.termVar^,
      cosi = erg.cosi^,
      sini = erg.sini^
    DO
      <* ASSERT NUMBER(variable) = NV *>
      vVar := variable;

      (* If a vertex is variable, its term is variable: *)
      FOR v := 0 TO NV-1 DO 
        IF cosi[v] # 0.0d0 OR sini[v] # 0.0d0 THEN
          <* ASSERT vertex[v].exists *>
          termVar[v] := vVar[v];
          WITH ov = out[v] DO
            VAR a: Arc := ov;
            BEGIN 
              REPEAT
                WITH 
                  e = NARROW(a.edge, Triang.Edge),
                  u = OrgV(Sym(a)), f = LeftF(a)
                DO
                  <* ASSERT f.exists AND e.exists AND u.exists *>
                  termVar[v] := termVar[v] OR vVar[u.num]
                END;
                a := Onext(a)
              UNTIL a = ov;
            END
          END
        ELSE
          termVar[v] := FALSE;
        END;
      END;
    END
  END DefVar;

PROCEDURE Eval(
    erg: T; 
    READONLY c: Coords; 
    VAR e: LONGREAL; 
    grad: BOOLEAN;
    VAR eDc: Gradient;
  ) =
  BEGIN
    WITH
      NV = erg.top.NV,
      out = erg.top.out^,
      vVar = erg.vVar^,
      termVar = erg.termVar^,
      cosi = erg.cosi^,
      sini = erg.sini^,
      K = erg.K
    DO

      PROCEDURE AddTerm(
          kv: CARDINAL;
          ov: Arc; 
          cosiv, siniv: LONGREAL;
        ) =
        (*
          Adds to "e" the energy term corresponding to 
          vertex "vertex[kv]", given one of its outgoing arcs "ov"
          and the cosine and sine of the ideal angle.
          
          Also adds to "eDc" the gradient terms that come from
          this energy term.
        *)

        PROCEDURE DoAddTerm() =
          VAR nv, eDnv: LR3.T;
          BEGIN
            ComputeVertexNormal(nv);
            AddSubTerms(nv, eDnv);
            IF grad THEN DistributeTorque(eDnv) END
          END DoAddTerm;

        PROCEDURE ComputeVertexNormal(VAR nv: LR3.T) =
          (* 
            Computes a (non-normalized) vector "nv" approximately
            perpendicular to the surface at "v", as the sum of "Cross(a,b)"
            for consecutive neighbors "a,b".
          *)
          VAR aa: Arc := ov; 
              ka: CARDINAL := OrgV(Sym(aa)).num;
          BEGIN
            nv := LR3.T{0.0d0, ..};
            REPEAT 
              WITH bb = Onext(aa), kb = OrgV(Sym(bb)).num DO
                nv := LR3.Add(nv, LR3Extras.Cross(c[ka], c[kb]));
                aa := bb; ka := kb
              END
            UNTIL aa = ov;
          END ComputeVertexNormal;
          
        PROCEDURE DistributeTorque (READONLY eDnv: LR3.T) =
          (* 
            Adds to "eDc[u]" the terms due to the influence of 
            "u" on the vertex normal "nv", given the influence "eDnv" 
            of "nv" on "e", where "u" ranges over the neighbors of "v". *)
          VAR aa: Arc := ov; 
              ka: CARDINAL := OrgV(Sym(aa)).num;
          BEGIN
            REPEAT 
              WITH bb = Onext(aa), kb = OrgV(Sym(bb)).num DO
                IF vVar[ka] THEN
                  eDc[ka] := LR3.Add(eDc[ka], LR3Extras.Cross(c[kb], eDnv))
                END;
                IF vVar[kb] THEN
                  eDc[kb] := LR3.Add(eDc[kb], LR3Extras.Cross(eDnv, c[ka]))
                END;
                aa := bb; ka := kb
              END
            UNTIL aa = ov
          END DistributeTorque;

        PROCEDURE AddSubTerms(READONLY nv: LR3.T; VAR eDnv: LR3.T) =
          (*
            Adds to "e" the energies for all pairs of consecutive
            edges out of "v", given the local (un-normalized) normal "nv".
            Also adds to "eDc" the corresponding gradient
            terms, and to "eDnv" the gradient relative to "nv".
          *)
           
          PROCEDURE DoAddSubTerms() =
            VAR aa: Arc := ov; 
                ka: CARDINAL := OrgV(Sym(aa)).num;
            BEGIN
              eDnv := LR3.T{0.0d0, ..};
              REPEAT 
                WITH bb = Onext(aa), kb = OrgV(Sym(bb)).num DO
                  AddSubTerm(c[kv], c[ka], c[kb], eDc[kv], eDc[ka], eDc[kb]);
                  aa := bb; ka := kb
                END;
              UNTIL aa = ov
            END DoAddSubTerms;

          PROCEDURE AddSubTerm(READONLY v, a, b: LR3.T; VAR eDv, eDa, eDb: LR3.T) =
            (*
              Adds to "e" the energy sub-term correponting to the consecutive
              edges "(v,a)" and "(v,b)".

              Also adds to "eDv", "eDa", and "eDb" the gradients of that
              sub-term relative to those arguments; and to the global
              "eDnv" the gradient of the sub-term relative to "nv".
            *)
            VAR et: LONGREAL;
                etDav, etDbv, etDnv: LR3.T;
            BEGIN
              WITH
                av = LR3.Sub(a, v),
                bv = LR3.Sub(b, v)
              DO
                ComputeSubTermFromSpokesAndNormal(av, bv, nv, et, etDav, etDbv, etDnv);
                e := e + et;
                IF grad THEN
                  eDa := LR3.Add(eDa, etDav);
                  eDb := LR3.Add(eDb, etDbv);
                  eDv := LR3.Sub(eDv, LR3.Add(etDav, etDbv));
                  eDnv := LR3.Add(eDnv, etDnv);
                END
              END
            END AddSubTerm;

          PROCEDURE ComputeSubTermFromSpokesAndNormal (
              READONLY av, bv, nv: LR3.T; 
              VAR et: LONGREAL; 
              VAR etDav, etDbv, etDnv: LR3.T;
            ) =
            (*
              Returns (does not add) in "et" the energy sub-term 
              for a pair of consecutive edges with vectors "av, bv"
              out of a vertex "v", given the vertex normal "nv". 
              
              Also, if "grad" is TRUE, returns (does not add) 
              in "etDav", "etDbv" and "etDnv" the gradients of the
              sub-term relative to the variable parameters.
            *)
            VAR etDr, etDs: LR3.T;
            BEGIN
              WITH
                r = LR3Extras.Cross(nv, av),
                s = LR3Extras.Cross(nv, bv)
              DO
                ComputeSubTermFromProjectedVectors(r, s, nv, et, etDr, etDs);
                IF grad THEN
                  etDav := LR3Extras.Cross(etDr, nv);
                  etDbv := LR3Extras.Cross(etDs, nv);
                  etDnv := LR3.Add(
                    LR3Extras.Cross(av, etDr),
                    LR3Extras.Cross(bv, etDs)
                  )
                END
              END
            END ComputeSubTermFromSpokesAndNormal;
            
          PROCEDURE ComputeSubTermFromProjectedVectors(
              READONLY r, s, nv: LR3.T; 
              VAR et: LONGREAL; 
              VAR etDr, etDs: LR3.T;
            ) =
            (*
              Returns (does not add) in "et" the energy sub-term 
              for a pair of consecutive edges, given the
              (scaled) projections "r, s" of their vectors 
              on the nominal tangent plane, and the surface normal "nv"
              
              Also returns (does not add) in "etDr" and "etDs"
              the gradients of the sub-term relative to "r" 
              and "s".  The gradient relative to "nv" is not
              returned because it is used only to get the 
              sign of the angle.
            
            *)

            PROCEDURE Sign(y: LONGREAL): LONGREAL = 
              BEGIN 
                IF y < 0.0d0 THEN RETURN 1.0d0 ELSE RETURN -1.0d0 END
              END Sign;

            BEGIN
              WITH
                m = LR3.Norm(r),
                n = LR3.Norm(s),
                mn = m * n,
                cmn = LR3.Dot(r, s),
                x = LR3Extras.Cross(r, s),
                sign = Sign(LR3.Dot(x, nv)),
                smn = sign * LR3.Norm(x)
              DO
                IF mn # 0.0d0 THEN
                  WITH
                    cosa = cmn/mn,
                    sina = smn/mn
                  DO
                    VAR etDcosa, etDsina: LONGREAL;
                    BEGIN
                      ComputeSubTermFromCosAndSin(cosa, sina, et, etDcosa, etDsina);
                      IF grad THEN
                        WITH
                          etDcmn= etDcosa/mn,
                          etDmn = - etDcosa*cosa/mn - etDsina*sina/mn,
                          etDm = etDmn * n,
                          etDn = etDmn * m
                        DO
                          etDr := LR3.Mix(etDcmn, s, etDm/m, r);
                          etDs := LR3.Mix(etDcmn, r, etDn/n, s);
                          IF smn # 0.0d0 THEN
                            WITH
                              etDsmn= etDsina/mn,
                              etDx = LR3.Scale(etDsmn/smn, x)
                            DO
                              etDr := LR3.Add(etDr, LR3Extras.Cross(s, etDx));
                              etDs := LR3.Add(etDs, LR3Extras.Cross(etDx, r));
                            END
                          END
                        END
                      END
                    END
                  END
                END
              END
            END ComputeSubTermFromProjectedVectors;

          PROCEDURE ComputeSubTermFromCosAndSin(
              cosa, sina: LONGREAL;
              VAR et: LONGREAL;
              VAR etDcosa, etDsina: LONGREAL;
            ) =
            (*
              Returns in "et" the angular strain energy from the 
              cosine and sine "cosa", "sina" of the actual projected 
              angle; and in "etDcosa, etDsina"
              the gradient of "et" relative to "cosa" and "sina".
            *)
            BEGIN
              WITH
                strain = 1.0d0 - (cosa*cosiv + sina*siniv)
              DO
                et := K * strain;
                IF grad THEN
                  WITH
                    etDstrain = K
                  DO
                    etDcosa := - etDstrain * cosiv;
                    etDsina := - etDstrain * siniv
                  END
                END
              END
            END ComputeSubTermFromCosAndSin;
        
          BEGIN
            DoAddSubTerms()
          END AddSubTerms;
        
        BEGIN
          DoAddTerm()
        END AddTerm;

      BEGIN
        e := 0.0d0;
        FOR v := 0 TO NV-1 DO eDc[v] := LR3.T{0.0d0, ..} END;
        FOR v := 0 TO NV-1 DO
          IF termVar[v] THEN AddTerm(v, out[v], cosi[v], sini[v]) END;
        END
      END;
    END
  END Eval;

PROCEDURE Name(<*UNUSED*> erg: T): TEXT =
  BEGIN
    RETURN "Corner()"
  END Name;

BEGIN
END CornerEnergy.
