MODULE ExcenEnergy;
IMPORT LR3, Triang; 
FROM Oct IMPORT Sym;
FROM Triang IMPORT OrgV, Topology;
FROM Energy IMPORT Coords, Gradient;

REVEAL
  T = Public BRANDED OBJECT
      top: Topology;
      K: LONGREAL;                     (* Energy normalization factor *)
      vVar: REF ARRAY OF BOOLEAN;      (* Tells which vertices are variable *)
      termVar: REF ARRAY OF BOOLEAN;   (* Tells which terms are variable *)
      deg: REF ARRAY OF CARDINAL;      (* Effective degree of each vertex *)
      org, dst: REF ARRAY OF CARDINAL; (* Directed edges of computed terms *)
      NP: CARDINAL;                    (* Number of edges in "org", "dst" *)
      sum: REF ARRAY OF LR3.T;         (* (Work) Sum of effective neighbors *)
      eDsum: REF ARRAY OF LR3.T;       (* (Work) Gradient of "e" rel "sum" *)
    OVERRIDES
      init := Init;
      defTop := DefTop;
      defVar := DefVar;
      eval := Eval;
      name := Name;
    END;
    
CONST
  Zero = LR3.T{0.0d0, 0.0d0, 0.0d0};
    
PROCEDURE Init(erg: T): T =
  BEGIN
    RETURN erg
  END Init;

PROCEDURE DefTop(erg: T; READONLY top: Topology) =
  BEGIN
    erg.top := top;
    erg.K := FLOAT(top.NV, LONGREAL)/64.0d0;
    erg.vVar := NEW(REF ARRAY OF BOOLEAN, top.NV);
    erg.termVar := NEW(REF ARRAY OF BOOLEAN, top.NV);
    erg.deg := NEW(REF ARRAY OF CARDINAL, top.NV);
    erg.org := NEW(REF ARRAY OF CARDINAL, 2*top.NE);
    erg.dst := NEW(REF ARRAY OF CARDINAL, 2*top.NE);
    erg.sum := NEW(REF ARRAY OF LR3.T, top.NV);
    erg.eDsum := NEW(REF ARRAY OF LR3.T, top.NV);
    (* In case the client forgets to call "defVar": *)
    erg.NP := 0;
    FOR i := 0 TO top.NV-1 DO 
      erg.vVar[i] := FALSE; erg.termVar[i] := FALSE
    END;
  END DefTop;
  
PROCEDURE DefVar(erg: T; READONLY variable: ARRAY OF BOOLEAN) =
  BEGIN
    (*
      Decide which terms are variable, and compute 
      effective degrees "erg.deg[v]".

      There is a term for each existing vertex. 
      The term of "v" is variable if "v" is variable
      or has some effective neighbor that is variable.
    *)
    WITH
      NP = erg.NP,
      NV = erg.top.NV,
      NE = erg.top.NE,
      vertex = erg.top.vertex^,
      edge = erg.top.edge^,
      vVar = erg.vVar^,
      termVar = erg.termVar^,
      org = erg.org^,
      dst = erg.dst^,
      deg = erg.deg^
    DO
      <* ASSERT NUMBER(variable) = NV *>
      vVar := variable;

      (* If a vertex is vVar, its term is vVar: *)
      FOR v := 0 TO NV-1 DO 
        termVar[v] := vVar[v] AND vertex[v].exists ; 
        deg[v] := 0
      END;    

      (* Enumerate effective edges and accumulate efective degrees. *)
      (* Also set "termVar[v]" if "v" has a vVar effective neighbor. *)
      FOR i := 0 TO NE-1 DO 
        WITH
          a = edge[i], e = NARROW(a.edge, Triang.Edge)
        DO
          IF e.exists THEN
            WITH
              u = OrgV(a),      un = u.num,
              v = OrgV(Sym(a)), vn = v.num
            DO
              <* ASSERT u.exists AND v.exists *>
              INC(deg[un]); INC(deg[vn]);
              termVar[un] := termVar[un] OR vVar[vn];
              termVar[vn] := termVar[vn] OR vVar[un];
            END
          END;
        END;
      END;

      (* A vertex with no effective neighbors has energy zero by definition: *)
      FOR v := 0 TO NV-1 DO 
        IF deg[v] = 0 THEN termVar[v] := FALSE END
      END;    

      (* Now collect all efective directed edges "(org[k], dst[k])" *)
      (*   whose origin "org[k]" is variable: *) 
      NP := 0;
      FOR i := 0 TO NE-1 DO 
        WITH
          a = edge[i], e = NARROW(a.edge, Triang.Edge)
        DO
          IF e.exists THEN
            WITH
              u = OrgV(a),      un = u.num,
              v = OrgV(Sym(a)), vn = v.num
            DO
              IF termVar[un] THEN org[NP] := un; dst[NP] := vn; INC(NP) END;
              IF termVar[vn] THEN org[NP] := vn; dst[NP] := un; INC(NP) END;
            END
          END;
        END;
      END;
    END
  END DefVar;
  
PROCEDURE Eval(
    erg: T; 
    READONLY c: Coords; 
    VAR e: LONGREAL; 
    grad: BOOLEAN;
    VAR eDc: Gradient;
  ) =
  BEGIN
    WITH
      NP = erg.NP,
      NV = erg.top.NV,
      vVar = erg.vVar^,
      termVar = erg.termVar^,
      deg = erg.deg^,
      org = erg.org^,
      dst = erg.dst^,
      sum = erg.sum^,
      eDsum = erg.eDsum^,
      K = erg.K
    DO

      PROCEDURE Compute_sum () =
        (* Computes "sum[v]" for all relevant terms "v" *)
        BEGIN
          FOR v := 0 TO NV-1 DO 
            IF termVar[v] THEN sum[v] := Zero END
          END;    
          FOR i := 0 TO NP-1 DO 
            WITH v = org[i], u = dst[i] DO
              <* ASSERT termVar[v] *>
              sum[v] := LR3.Add(sum[v], c[u])
            END
          END 
        END Compute_sum;

      PROCEDURE EvalTerm (
          READONLY cv, sumv: LR3.T; 
          degv: CARDINAL;
          VAR ev: LONGREAL; 
          VAR evDcv, evDsumv: LR3.T;
        ) =
        (*
          Computes the energy term "ev" for a vertex at "cv",
          given the sum of its relevant neighbors "sumv".
          Returns also the gradients "evDcv" and "evDsumv" of
          "ev" relative to "cv" and "sumv".
        *)
        BEGIN
          WITH
            f = 1.0d0/FLOAT(degv, LONGREAL),
            barv = LR3.Scale(f, sumv),
            rv = LR3.Sub(cv, barv)
          DO
            ev := K * LR3.NormSqr(rv);
            IF grad THEN
              evDcv := LR3.Scale(2.0d0 * K, rv);
              evDsumv := LR3.Scale(-2.0d0 * f * K, rv)
            ELSE
              evDcv := Zero;
              evDsumv := Zero
            END
          END
        END EvalTerm;

      PROCEDURE Distribute_eDsum () =
        (*
          Adds to the derivatives "eDc[u]" the indirect terms
          represented by "eDsum[v]", for all computed terms "v" 
          that have "u" as a relevant neighbor. 
        *)
        BEGIN
          FOR i := 0 TO NP-1 DO 
            WITH v = org[i], u = dst[i] DO
              IF vVar[u] THEN
                eDc[u] := LR3.Add(eDc[u], eDsum[v])
              END
            END
          END;  
        END Distribute_eDsum;

      VAR ev: LONGREAL;
      BEGIN
        Compute_sum();
        e := 0.0d0;
        FOR v := 0 TO NV-1 DO
          IF termVar[v] THEN
            EvalTerm(c[v], sum[v], deg[v], ev, eDc[v], eDsum[v]);
            e := e + ev
          ELSE
            eDc[v] := Zero;
          END;
        END;
        IF grad THEN Distribute_eDsum() END
      END;
    END
  END Eval;

PROCEDURE Name(<*UNUSED*> erg: T): TEXT =
  BEGIN
    RETURN "Excen()"
  END Name;

BEGIN
END ExcenEnergy.


