MODULE BrentUniMin;

FROM UniMin IMPORT EvalProc, CheckProc, ComputeDerivatives;
IMPORT Wr, Thread, Fmt;
FROM Stdio IMPORT stderr;

REVEAL
  T = Public BRANDED OBJECT
    OVERRIDES
      minimize := Minimize;
      needsDerivative := NeedsDerivative;
      name := Name;
    END;
    
PROCEDURE NeedsDerivative(<*UNUSED*> bum: T): BOOLEAN =
  BEGIN
    RETURN FALSE   
  END NeedsDerivative;

PROCEDURE Name(<*UNUSED*> bum: T): TEXT =
  BEGIN
    RETURN "BrentUniMin"
  END  Name;
  
PROCEDURE Minimize(
    bum: T;
    eval: EvalProc;
    VAR x: LONGREAL;
    VAR fx: LONGREAL;
    VAR dfx: LONGREAL;
    tol: LONGREAL;
    dist: LONGREAL;
    a: LONGREAL;
    b: LONGREAL;
    check: CheckProc;
  ) =

  CONST 
    Eps = 1.0d-6;                     (* Square root of relative machine precision. *)
    Phi = 1.61803398874989484821d0;   (* Golden ratio *)
    IhpSqr = 0.3819660112501051518d0; (* 1/Phi^2 *)

  VAR 
    v, fv: LONGREAL;  (* Third-smallest point of current triple *)
    w, fw: LONGREAL;  (* Second-smallest point of current triple *)
    u, fu: LONGREAL;  (* Splitting point *)
    d, e, r: LONGREAL;     (* Step ssizes *)
    dp, ddp, q: LONGREAL;  (* Estimated derivatives at "x" *)
    xOld: LONGREAL := x;   (* Previous value of "x" *)
    dfu: LONGREAL;         (* Derivative at "u" (returned but ignored). *)

  PROCEDURE PrintStatus () =
    BEGIN
      PrintNewLine();
      PrintAbscissa("a", a);
      PrintPoint("v", v, fv);
      PrintPoint("x", x, fx);
      PrintPoint("w", w, fw);
      PrintAbscissa("b", b);
      PrintNewLine();
    END PrintStatus;

  BEGIN
  
    <* ASSERT a <= x *>
    <* ASSERT x <= b *>

    WITH
      tol3 = tol/3.0d0,
      Debug = bum.debug
    DO
      w := x; fw := fx;
      v := x; fv := fx;
      e := 0.0d0;
      LOOP
        <* ASSERT fx <= fw *>
        <* ASSERT fw <= fv *>
        WITH
          aInf = (a = FIRST(LONGREAL)),
          bInf = (b = LAST(LONGREAL)),
          tol1 = Eps * ABS(x) + tol3,
          tol2 = 2.0d0 * tol1
        DO
          IF aInf THEN <* ASSERT x <= w AND w <= v *> END;
          IF bInf THEN <* ASSERT x >= w AND w >= v *> END;

          IF Debug THEN PrintStatus() END;

          (* check stopping criterion: *)
          IF NOT aInf AND NOT bInf AND MAX(x-a, b-x) <= tol2 THEN EXIT END;

          IF ABS(e) > tol1 THEN 
            (* Fit parabola through "v", "x", "w"; *)
            ComputeDerivatives(v, fv, w, fw, x, fx, dp, ddp, q);
            r := e; e := d
          ELSE
            dp := 0.0d0; ddp := 0.0d0; q := 0.0d0;
            r := 0.0d0;
          END;

          IF x # xOld THEN
            IF check # NIL THEN
              IF check(a, b, x, fx, dp, ddp, q) THEN EXIT END
            END;
            xOld := x;
          END;

          (* Compute the displacement "d" from "x" to the next probe: *)
          IF ddp < 0.0d0 THEN dp := -dp; ddp := -ddp END;
          IF ABS(dp) < ABS(0.5d0*ddp*r) 
          AND (aInf OR dp < ddp*(x - a)) 
          AND (bInf OR dp > ddp*(x - b)) THEN 
            (* A parabolic-interpolation step: *)
            IF Debug THEN Message("parabolic step") END;
            d := -dp/ddp;

            (* "f" must not be evaluated too close to "a" or "b": *)
            u := x + d;
            IF NOT aInf AND u - a < tol2 THEN 
              d := tol1
            ELSIF NOT bInf AND b - u < tol2 THEN
              d := -tol1
            END;
          ELSIF bInf OR aInf THEN
            (* A golden-section extrapolation: *)
            IF Debug THEN Message("golden extension") END;
            IF x # w THEN e := x - w ELSIF x # v THEN e := x - v ELSE e := 0.0d0 END;
            IF e = 0.0d0 THEN d := Phi*dist ELSE d := Phi*e END;
            IF ABS(d) < tol1 THEN 
              IF d < 0.0d0 THEN d := -tol1 ELSE d := tol1 END
            END
          ELSE
            (* A golden-section interpolation: *)
            IF Debug THEN Message("golden section") END;
            IF b - x > x - a THEN e := b - x ELSE e := a - x END;
            d := IhpSqr * e;
          END;

          (* f must not be evaluated too close to x: *)
          IF ABS(d) >= tol1 THEN
            u := x + d
          ELSIF d > 0.0d0 THEN
            u := x + tol1
          ELSE
            u := x - tol1
          END;  
          eval(u, fu, dfu);

          (* update  a, b: *)
          IF fu >= fx THEN
            IF u < x THEN a := u ELSE b := u END
          END;
          IF fu <= fx THEN
            IF u < x THEN b := x ELSE a := x END;
          END;

          (* Update "x", "v", "w": *)
          IF fu <= fx THEN
            v := w; fv := fw;
            w := x; fw := fx;
            x := u; fx := fu; dfx := dfu
          ELSIF fu <= fw THEN
            <* ASSERT w # x *>
            v := w; fv := fw;
            w := u; fw := fu;
          ELSIF fu <= fv THEN
            <* ASSERT v # x *>
            IF w = x THEN 
              w := u; fw := fu
            ELSE
              v := u; fv := fu
            END
          ELSIF v = w THEN
            v := u; fv:= fu
          END;
        END
      END
    END
  END Minimize;

PROCEDURE Message (msg: TEXT) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, "\n");
  END Message;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

PROCEDURE PrintAbscissa (msg: TEXT; u: LONGREAL) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN 
    Wr.PutText(stderr, "    ");
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(u), 12));
    Wr.PutText(stderr, "\n");
  END PrintAbscissa;

PROCEDURE PrintPoint (msg: TEXT; u, fu: LONGREAL) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN 
    Wr.PutText(stderr, "    ");
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(u), 12));
    Wr.PutText(stderr, " f" & msg & " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(fu), 12));
    Wr.PutText(stderr, "\n");
  END PrintPoint;

BEGIN
END BrentUniMin.
