MODULE CoordMinimizer;

IMPORT LRN, Math, UniMin, Minimizer, BrentUniMin; 
FROM Minimizer IMPORT 
  Point, Coord, Gradient, Length, Value, Slope, EvalProc, TraceProc;

IMPORT Wr, Fmt, Thread;
FROM Stdio IMPORT stderr;

REVEAL
  T = Public BRANDED OBJECT
      budget: CARDINAL := 5;  (* Evals per coordinate *)
      NC: CARDINAL;           (* Number of client variables *)
      u: UniMin.T;            (* Univariate minimizer *)
      ox: REF Point;          (* (Work) Position at end of last diagonal minimization *)
      wx: REF Point;          (* (Work) Current position *)
      wg: REF Gradient;       (* (Work) Gradient storage *)
      dx: REF Point;          (* (Work) Step size per coordinate *)
    OVERRIDES
      name := Name;
      setUniMin := SetUniMin;
      setBudget := SetBudget;
      setSize := SetSize;
      minimize := Minimize;
      needsGradient := NeedsGradient;
    END;
    
PROCEDURE Name(m: T): TEXT =
  BEGIN
    IF m.u = NIL THEN m.u := NEW(BrentUniMin.T) END;
    RETURN "CoordMinimizer("
     & "uniMin = " & m.u.name() 
     & ", budget = " & Fmt.Int(m.budget)
     & ")"
  END  Name;

PROCEDURE SetUniMin(m: T; u: UniMin.T): T =
  BEGIN
    m.u := u;
    RETURN m
  END SetUniMin;
    
PROCEDURE SetBudget(m: T; n: CARDINAL): T =
  BEGIN
    m.budget := n;
    RETURN m
  END SetBudget;
    
PROCEDURE SetSize(m: T; n: CARDINAL): Minimizer.T =
  BEGIN
    IF m.u = NIL THEN m.u := NEW(BrentUniMin.T) END;
    m.NC := n;
    m.wg := NEW(REF Gradient, m.NC);
    m.wx := NEW(REF Point, m.NC);
    m.dx := NEW(REF Point, m.NC);
    m.ox := NEW(REF Point, m.NC);
    RETURN m
  END SetSize;
  
PROCEDURE NeedsGradient(m: T): BOOLEAN =
  BEGIN
    IF m.u = NIL THEN m.u := NEW(BrentUniMin.T) END;
    RETURN m.u.needsDerivative()    
  END NeedsGradient;

PROCEDURE Minimize(
    m: T;
    eval: EvalProc;
    VAR x: Point;
    VAR f: Value;
    VAR g: Gradient;
    dist: Length;
    flat: Slope;
    tol: Length;
    maxEvals: CARDINAL;
    trace: TraceProc;
  ) =
  VAR totEvals: CARDINAL := 0; (* Total function evaluations *)
  BEGIN
    WITH 
      debug = m.debug,
      NC = m.NC,
      sqrtNC = Math.sqrt(FLOAT(NC, LONGREAL)),
      budget = m.budget,
      wg = m.wg^,
      wx = m.wx^,
      dx = m.dx^,
      ox = m.ox^
    DO
      <* ASSERT NUMBER(x) = NC *>
      
      VAR which: CARDINAL;       (* Which coordinate is being minimized *)
          clock: CARDINAL;       (* Counts axial mins since last diagonal min *)
          nEvals: CARDINAL := 0; (* Function evaluations of this coordinate *)
      
      PROCEDURE EvalAxial(z: LONGREAL; VAR fz, dfz: LONGREAL) =
        (*
          Called by "m.u.minimize" to compute the function's VALUE
          for point "wx" when coordinate "wx[which]" is set to "z".
          Also updates the best point "(x,f,g)" when "fz" 
          becomes less than "f".
        *)
        BEGIN
          wx[which] := z;
          eval(wx, fz, wg); 
          dfz := wg[which];
          INC(nEvals); INC(totEvals);
          IF debug THEN 
            Debug("EvalAxial: wx = ", wx);
            Debug("EvalAxial: fz = ", LRN.T{fz});
            Debug("EvalAxial: wg = ", wg);
            PrintNewLine()
          END;
          IF fz < f THEN 
            x := wx; f := fz; g := wg;
            IF trace # NIL THEN trace(wx, fz, wg) END;
          END;
        END EvalAxial;
        
      PROCEDURE EvalDiagonal(z: LONGREAL; VAR fz, dfz: LONGREAL) =
        (*
          Called by "m.u.minimize" to compute the function's value
          at a point "wx = ox + z * dx".
        *)
        BEGIN
          FOR i := 0 TO NC-1 DO 
            wx[i] := ox[i] + z * dx[i]
          END;
          eval(wx, fz, wg); 
          dfz := LRN.Dot(wg, dx);
          INC(nEvals); INC(totEvals);
          IF debug THEN 
            Debug("EvalDiagonal: wx = ", wx);
            Debug("EvalDiagonal: fz = ", LRN.T{fz});
            Debug("EvalDiagonal: wg = ", wg);
            PrintNewLine()
          END;
          IF fz < f THEN 
            x := wx; f := fz; g := wg;
            IF trace # NIL THEN trace(wx, fz, wg) END;
          END;
        END EvalDiagonal;
        
      PROCEDURE Check(
          <*UNUSED*> a, b: LONGREAL;
          <*UNUSED*> x, fx: LONGREAL;
          <*UNUSED*> dfx, ddfx, q: LONGREAL; 
        ): BOOLEAN =
        (*
          Called by "m.u.minimize" to decide whether to stop: *)
        BEGIN
          RETURN nEvals >= budget
        END Check;
     
      VAR t: Coord;         (* Parameter for univariate minimization *)
          ft: Value;        (* Function value along same *)
          dft: Slope;       (* Derivative of "ft" rel to "t" (may be used by "m.u") *)
          tOld: Coord;      (* Value of "t" before axial min *)
          ftOld: Value;     (* Value of "ft" before axial min *)
          fBase: Value;     (* Value of "f" after first axial min *)
      BEGIN
        which := 0;
        clock := 0;
        FOR i := 0 TO LAST(dx) DO dx[i] := dist/sqrtNC END;
        WHILE totEvals < maxEvals DO
          IF clock <= NC THEN
            (* Axial minimization *)
            t := x[which]; ft := f; dft := g[which];
            tOld := t; ftOld := ft;
            wx := x;
            nEvals := 0;
            m.u.minimize(
              eval := EvalAxial,
              x := t, fx := ft, dfx := dft,
              dist := ABS(dx[which]) + 0.1d0 * tol,
              tol := tol, 
              a := FIRST(LONGREAL), b := LAST(LONGREAL),
              check := Check
            );
            <* ASSERT ft = f *>
            IF clock = 0 THEN fBase := f END;
            dx[which] := x[which] - tOld;
            INC(clock)
          ELSE
            (* Diagonal optimization *)
            WITH
              dxNorm = LRN.Norm(dx)
            DO
              IF debug THEN 
                Debug("diagonal step: dxNorm = ", LRN.T{dxNorm});
              END;
              IF dxNorm <= 0.05d0 * tol THEN RETURN END;
              LRN.Scale(1.0d0/dxNorm, dx);
              t := 0.0d0; ft := f; dft := LRN.Dot(dx, g);
              ox := x;
              nEvals := 0;
              IF debug THEN 
                Debug("diagonal step: ox = ", ox);
                Debug("diagonal step: dx = ", dx);
                PrintNewLine();
              END;
              m.u.minimize(
                eval := EvalDiagonal,
                x := t, fx := ft, dfx := dft,
                dist := dxNorm + 0.1d0 * tol,
                tol := tol, 
                a := FIRST(LONGREAL), b := LAST(LONGREAL),
                check := Check
              );
              <* ASSERT ft = f *>
              WITH
                slope = ABS((fBase - f)/(dxNorm + t))
              DO
                IF debug THEN 
                  Debug("diagonal step: t, f, slope = ", LRN.T{t, f, slope})
                END;
                IF slope <= flat THEN RETURN END
              END;
              LRN.Scale(dxNorm, dx); 
            END;
            clock := 0;
          END;
          which := (which + 1) MOD NC;
        END
      END
    END
  END  Minimize;

PROCEDURE Debug(msg: TEXT; READONLY x: LRN.T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "CoordMinimizer ");
    Wr.PutText(stderr, msg);
    FOR i := 0 TO LAST(x) DO
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(x[i], Fmt.Style.Sci, 6), 15))
    END;
    Wr.PutText(stderr, "\n")
  END Debug;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

BEGIN
END CoordMinimizer.
