MODULE DownMinimizer;

IMPORT LRN, Minimizer; 
FROM Minimizer IMPORT 
  Point, Gradient, Length, Value, Slope, EvalProc, TraceProc;

IMPORT Wr, Fmt, Thread;
FROM Stdio IMPORT stderr;

REVEAL
  T = Public BRANDED OBJECT
      alpha: LONGREAL := 3.06956450765297882d0;
      beta: LONGREAL := 0.5d0;
      setSizeCalled: BOOLEAN := FALSE;
      NC: CARDINAL;     (* Number of client variables *)
      xn: REF Point;    (* (Work) probe point *)
      gn: REF Gradient; (* (Work) its gradient *)
    OVERRIDES
      name := Name;
      setAlphaBeta := SetAlphaBeta;
      setSize := SetSize;
      minimize := Minimize;
      needsGradient := NeedsGradient;
    END;
    
PROCEDURE Name(m: T): TEXT =
  BEGIN
    RETURN "DownMinimizer("
     & "alpha = " & Fmt.LongReal(m.alpha)
     & ", beta = " & Fmt.LongReal(m.beta)
     & ")"
  END  Name;

PROCEDURE SetAlphaBeta(m: T; alpha, beta: LONGREAL): T =
  BEGIN
    <* ASSERT alpha > 1.0d0 *>
    <* ASSERT beta < 1.0d0 AND beta > 0.0d0 *>
    m.alpha := alpha;
    m.beta := beta;
    RETURN m
  END SetAlphaBeta;
    
PROCEDURE SetSize(m: T; n: CARDINAL): Minimizer.T =
  BEGIN
    m.setSizeCalled := TRUE;
    m.NC := n;
    m.xn := NEW(REF Point, m.NC);
    m.gn := NEW(REF Gradient, m.NC);
    RETURN m
  END SetSize;

PROCEDURE NeedsGradient(<*UNUSED*> m: T): BOOLEAN =
  BEGIN
    RETURN TRUE   
  END NeedsGradient;

PROCEDURE Minimize(
    m: T;
    eval: EvalProc;
    VAR x: Point;
    VAR f: Value;
    VAR g: Gradient;
    dist: Length;
    flat: Slope;
    tol: Length;
    maxEvals: CARDINAL;
    trace: TraceProc;
  ) =
  VAR fn: Value;                    (* New function VALUE *)
      gNorm: Slope;                 (* Gradient norm at last "Diff" call *)
      nEvals: CARDINAL := 0;        (* Counts calls to the "eval" procedure. *)
      step: LONGREAL := dist;       (* Current step size *)
  BEGIN
    <* ASSERT m.setSizeCalled *>
    IF maxEvals <= 0 THEN RETURN END;
    WITH 
      debug = m.debug,
      alpha = m.alpha,
      beta = m.beta,
      NC = m.NC,
      xn = m.xn^,
      gn = m.gn^,
      minStep = tol,   (* Min step that can still be reduced *)
      maxStep = dist   (* Max step that can still be increased *)
    DO
      <* ASSERT NUMBER(x) = NC *>
      
      IF debug THEN 
        WITH gNorm = LRN.Norm(g) DO
          Debug("initial state:", LRN.T{});
          Debug("  f, gNorm, step  = ", LRN.T{f, gNorm, step}); 
          Debug("  x  = ", x); 
          Debug("  g  = ", g);
          PrintNewLine();
        END
      END;

      (* "x", "f", and "g" contain the best minimum seen so far. *)
      LOOP

        IF nEvals >= maxEvals THEN RETURN END;
        gNorm := LRN.Norm(g);
        IF gNorm <= flat THEN RETURN END;
        
        (* Compute new probe point: *)
        WITH s = step/gNorm DO
          FOR i := 0 TO NC-1 DO xn[i] := x[i] - s * g[i] END
        END;
        eval(xn, fn, gn); INC(nEvals);
        IF trace # NIL THEN trace(xn, fn, gn) END;

        IF debug THEN 
          WITH gnNorm = LRN.Norm(gn) DO
            Debug("state at nEvals = ", LRN.T{FLOAT(nEvals, LONGREAL)}); 
            Debug("  fn, gNorm, step = ", LRN.T{fn, gnNorm, step}); 
            Debug("  xn = ", xn); 
            Debug("  gn = ", gn);
          END
        END;

        (* Decide what to do: *)
        IF fn < f THEN
          (* Advance and increase step size: *)
          x := xn; f := fn; g := gn;
          IF step < maxStep THEN step := step * alpha END;
          IF debug THEN Debug("  accepted", LRN.T{}) END;
        ELSIF step <= tol THEN
          (* Give up: *)
          IF debug THEN Debug("  rejected and stopped", LRN.T{}) END;
          RETURN
        ELSE
          (* Reduce step size and retry: *)
          IF debug THEN Debug("  rejected", LRN.T{}) END;
          IF step > minStep THEN step := step * beta END;
        END;
        IF debug THEN PrintNewLine() END;
      END
    END
  END  Minimize;
  
PROCEDURE Debug(msg: TEXT; READONLY x: LRN.T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "DownMinimizer ");
    Wr.PutText(stderr, msg);
    FOR i := 0 TO LAST(x) DO
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(x[i], Fmt.Style.Sci, 6), 15))
    END;
    Wr.PutText(stderr, "\n")
  END Debug;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

BEGIN
END DownMinimizer.



