MODULE GradMinimizer;

IMPORT LRN, Integrator, RKF4Integrator, Minimizer; 
FROM Integrator IMPORT Time, Speed, State, Error, Velocity;
FROM Minimizer IMPORT 
  Point, Gradient, Length, Value, Slope, EvalProc, TraceProc;

IMPORT Wr, Fmt, Thread;
FROM Stdio IMPORT stderr;

REVEAL
  T = Public BRANDED OBJECT
      setSizeCalled: BOOLEAN := FALSE;
      NC: CARDINAL;     (* Number of client variables *)
      NS: CARDINAL;     (* Number of path state variables *)
      path: Path;       (* Path equations *)
      i: Integrator.T;  (* Path integrator *)
      s: REF State;     (* (Work) Path state *)
    OVERRIDES
      name := Name;
      setPath := SetPath;
      setIntegrator := SetIntegrator;
      setSize := SetSize;
      minimize := Minimize;
      needsGradient := NeedsGradient;
    END;
    
PROCEDURE Name(m: T): TEXT =
  BEGIN
    IF m.path = NIL THEN m.path := path[1]; m.setSizeCalled := FALSE END;
    IF m.i = NIL THEN m.i := NEW(RKF4Integrator.T); m.setSizeCalled := FALSE END;
    RETURN "GradMinimizer("
     & "integrator = " & m.i.name() 
     & ", path = {" & m.path.eqs() & "}"
     & ")"
  END  Name;

PROCEDURE SetPath(m: T; p: Path): T =
  BEGIN
    m.path := p;
    m.setSizeCalled := FALSE;
    RETURN m
  END SetPath;
    
PROCEDURE SetIntegrator(m: T; i: Integrator.T): T =
  BEGIN
    m.i := i;
    m.setSizeCalled := FALSE;
    RETURN m
  END SetIntegrator;
    
PROCEDURE SetSize(m: T; n: CARDINAL): Minimizer.T =
  BEGIN
    m.setSizeCalled := TRUE;
    IF m.path = NIL THEN m.path := path[1] END;
    IF m.i = NIL THEN m.i := NEW(RKF4Integrator.T) END;

    m.NC := n;
    m.NS := m.path.size(m.NC);

    m.s := NEW(REF State, m.NS);

    EVAL m.i.setSize(m.NS);
    RETURN m
  END SetSize;

PROCEDURE NeedsGradient(<*UNUSED*> m: T): BOOLEAN =
  BEGIN
    RETURN TRUE   
  END NeedsGradient;

PROCEDURE Minimize(
    m: T;
    eval: EvalProc;
    VAR x: Point;
    VAR f: Value;
    VAR g: Gradient;
    dist: Length;
    flat: Slope;
    tol: Length;
    maxEvals: CARDINAL;
    trace: TraceProc;
  ) =
  VAR gNorm: Slope;          (* Gradient norm at last "Diff" call *)
      nEvals: CARDINAL := 0; (* Counts calls to the "eval" procedure. *)
  BEGIN
    <* ASSERT m.setSizeCalled *>
    IF maxEvals <= 2 THEN RETURN END;
    WITH 
      debug = m.debug,
      NC = m.NC,
      NS = m.NS
    DO
      <* ASSERT NUMBER(x) = NC *>
      
      (* "x" and "f" contain the best minimum seen so far. *)
      
      PROCEDURE InitialState(VAR s: State): Time =
        (*
          Set "s" to the initial state and return the initial time step "dt".
          Assumes the gradient "g" and the function value "f" are
          have already been computed by the client.
        *)
        BEGIN
          WITH
            sx = SUBARRAY(s, 0, NC),
            su = SUBARRAY(s, NC, NS-NC)
          DO
            sx := x;
            gNorm := LRN.Norm(g);
            IF gNorm = 0.0d0 THEN
              LRN.Fill(su, 0.0d0);
              RETURN 0.0d0
            ELSE 
              WITH 
                speed = m.path.start(g, gNorm, su),
                dt = dist/speed
              DO
                IF debug THEN 
                  Debug("InitialState: t, speed, dt = ", LRN.T{0.0d0, speed, dt}); 
                  Debug("InitialState: s = ", s);
                  PrintNewLine();
                END;
                RETURN dt
              END
            END;
          END
        END InitialState;

      PROCEDURE Diff (* : Integrator.DiffProc *) (
          t: Time; READONLY s: State; VAR v: Velocity
        ) =
        (*
          Called by "Integrate" to compute the state's derivative
          "v" for state "s". Calls the client's "eval" procedure
          and maps it through the path equations.  Also saves in 
          "gNorm" the gradient magnitude and in "f" the function
          value, and updates "x" and "fMin".
        *)
        VAR sf: Value;
        BEGIN
          WITH
            sx = SUBARRAY(s, 0, NC),     vx = SUBARRAY(v, 0, NC),
            su = SUBARRAY(s, NC, NS-NC), vu = SUBARRAY(v, NC, NS-NC)
          DO
            eval(sx, sf, g); INC(nEvals);
            gNorm := LRN.Norm(g);
            IF sf < f THEN 
              x := sx; f := sf; 
            END;
            IF gNorm = 0.0d0 THEN
              LRN.Fill(v, 0.0d0)
            ELSE
              m.path.diff(g, gNorm, su, vx, vu);
            END;
            IF debug THEN 
              Debug("Diff: t = ", LRN.T{t}); 
              Debug("Diff: f, gNorm = ", LRN.T{f, gNorm}); 
              Debug("Diff: s = ", s); 
              Debug("Diff: v = ", v);
              Debug("Diff: g = ", g);
              PrintNewLine();
            END;
          END
        END Diff;

      PROCEDURE CheckState (* : Integrator.CheckStateProc *) (
          t: Time;
          READONLY s: State; 
          READONLY v: Velocity;
        ): Time =
        (*
          See the definition of "Integrator.CheckStateProc".
          Assumes "Diff" was just called for this state,
          so "gNorm" is the slope there.
        *)
        BEGIN
          WITH
            sx = SUBARRAY(s, 0, NC)
          DO
            IF trace # NIL THEN 
              trace(sx, f, g)
            END;
            IF gNorm <= flat OR nEvals >= maxEvals THEN
              RETURN t
            ELSE
              WITH
                vx = SUBARRAY(v, 0, NC),
                speed = LRN.Norm(vx)
              DO
                RETURN t + MIN(1.0d100, MAX(1.0d-100, dist/speed))
              END
            END
          END
        END CheckState;

      PROCEDURE CheckStep (* : Integrator.CheckStepProc *) (
          ta: Time;
          READONLY sa: State;
          <*UNUSED*> READONLY va: Velocity;
          tb: Time;
          VAR sb: State;
          READONLY er: Error;
        ): Time =
        (*
          See the definition of "Integrator.CheckStepProc".
        *)
        BEGIN
          IF nEvals > maxEvals THEN
            (* Stop right there, boy *)
            RETURN tb
          END;
          WITH 
            dt = tb - ta,
            xa = SUBARRAY(sa, 0, NC),
            xb = SUBARRAY(sb, 0, NC),
            xe = SUBARRAY(er, 0, NC),
            error = LRN.Norm(xe),
            speed = LRN.Dist(xa, xb)/dt,
            dtMin = MAX(1.0d-100,   0.001d0*tol/speed),
            dtMax = MIN(1.0d+100, 100.0d0*dist/speed),
            dtIdeal = m.i.computeStep(dt, error, 0.5d0 * tol),
            dtNew = MIN(dtMax, MAX(dtMin, dtIdeal))
          DO
            IF debug THEN 
              Debug("CheckStep: ta, speed = ", LRN.T{ta, speed});
              Debug("CheckStep: error/tol = ", LRN.T{error/tol});
              Debug("CheckStep: dtMin, dtMax = ", LRN.T{dtMin, dtMax});
              Debug("CheckStep: dt, dtNew = ", LRN.T{dt, dtNew});
              PrintNewLine()
            END;
            IF error < tol THEN
              (* Step was short enough, keep going *)
              RETURN tb + dtNew
            ELSE
              (* Redo step with new step length: *)
              RETURN ta + dtNew
            END
          END
        END CheckStep;

      BEGIN
        WITH
          s = m.s^,
          dtStart = InitialState(s)
        DO
          (* Follow the path: *)
          EVAL m.i.integrate(
            diff := Diff,
            t := 0.0d0,
            s := s,
            dt := dtStart,
            checkState := CheckState,
            checkStep := CheckStep
          );
        END 
      END

    END
  END  Minimize;
  
PROCEDURE Debug(msg: TEXT; READONLY x: LRN.T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "GradMinimizer ");
    Wr.PutText(stderr, msg);
    FOR i := 0 TO LAST(x) DO
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(x[i], Fmt.Style.Sci, 6), 15))
    END;
    Wr.PutText(stderr, "\n")
  END Debug;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

REVEAL 
  Path = PublicPath BRANDED OBJECT METHODS
      size(NC: CARDINAL): CARDINAL;
        (*
          Computes the number of path state variables from the number of
          function variables. 
        *)
      
      start(READONLY g: Gradient; gNorm: Slope; VAR u: LRN.T): Speed;
        (* 
          Computes the initial values of auxiliary variables.
          Returns the initial speed. Called only if "gNorm" is greater
          than "flat".
        *)

      diff(READONLY g: Gradient; gNorm: Slope; READONLY u: LRN.T; VAR xDt, uDt: LRN.T);
        (*
          Computes the state derivatives "xDt" and "uDt" from 
          the local gradient "g" and the auxiliary variables "u"
        *)
    END;
    

(*********************************************************************)

TYPE Path1 = Path OBJECT OVERRIDES 
      eqs := Eqs1; size := Size1; start := Start1; diff := Diff1;
    END;

PROCEDURE Eqs1(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN "dx/dt = -g/|g|"
  END Eqs1;
  
PROCEDURE Size1(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN NC END Size1;
  
PROCEDURE Start1(<*UNUSED*> p: Path; 
    <*UNUSED*> READONLY g: Gradient; <*UNUSED*> gNorm: Slope; <*UNUSED*> VAR u: LRN.T
  ): Speed =
  BEGIN RETURN 1.0d0 END Start1;
  
PROCEDURE Diff1(<*UNUSED*> p: Path; 
    READONLY g: Gradient; gNorm: Slope;
    <*UNUSED*> READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN
    xDt := g; LRN.Scale(-1.0d0/gNorm, xDt);
    <* ASSERT NUMBER(uDt) = 0 *>
  END Diff1;
  
(*********************************************************************)

TYPE Path2 = Path OBJECT OVERRIDES 
      eqs := Eqs2; size := Size2; start := Start2; diff := Diff2
    END;

PROCEDURE Eqs2(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN "dx/dt = -g"
  END Eqs2;
  
PROCEDURE Size2(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN NC END Size2;
  
PROCEDURE Start2(<*UNUSED*> p: Path;
    <*UNUSED*> READONLY g: Gradient; gNorm: Slope; <*UNUSED*> VAR u: LRN.T
  ): Speed =
  BEGIN RETURN gNorm END Start2;
  
PROCEDURE Diff2(<*UNUSED*> p: Path; 
    READONLY g: Gradient; <*UNUSED*> gNorm: Slope;
    <*UNUSED*> READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN 
    xDt := g; LRN.Neg(xDt);
    <* ASSERT NUMBER(uDt) = 0 *>
  END Diff2;
  
(*********************************************************************)

TYPE 
  Path3 = Path OBJECT OVERRIDES 
      eqs := Eqs3; size := Size3; start := Start3; diff := Diff3
    END;

PROCEDURE Eqs3(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN 
      "dx/dt = - ((g.u)/|g|)u" & "; " &
      "du/dt = - g/|g| - ((g.u)/|g|)u"
  END Eqs3;
  
PROCEDURE Size3(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN 2*NC END Size3;
  
PROCEDURE Start3(<*UNUSED*> p: Path;
    READONLY g: Gradient; gNorm: Slope; VAR u: LRN.T
  ): Speed =
  BEGIN 
    u := g; 
    LRN.Scale(-1.0d0/gNorm, u);
    RETURN 1.0d0
  END Start3;
  
PROCEDURE Diff3(<*UNUSED*> p: Path; 
    READONLY g: Gradient; gNorm: Slope;
    READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN
    WITH
      a = -LRN.Dot(g, u)/gNorm, 
      b = -1.0d0/gNorm
    DO
      FOR i := 0 TO LAST(g) DO
        xDt[i] := a * u[i];
        uDt[i] := b * g[i] - xDt[i]
      END
    END
  END Diff3;
  
(*********************************************************************)

TYPE 
  Path4 = Path OBJECT OVERRIDES 
      eqs := Eqs4; size := Size4; start := Start4; diff := Diff4
    END;

PROCEDURE Eqs4(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN 
      "dx/dt = - (1/|g|)((g.u)/(u.u))u" & "; " &
      "du/dt = - g - u"
  END Eqs4;
  
PROCEDURE Size4(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN 2*NC END Size4;
  
PROCEDURE Start4(<*UNUSED*> p: Path;
    READONLY g: Gradient; gNorm: Slope; VAR u: LRN.T
  ): Speed =
  BEGIN
    u := g;
    RETURN gNorm
  END Start4;
  
PROCEDURE Diff4(<*UNUSED*> p: Path; 
    READONLY g: Gradient; gNorm: Slope;
    READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN
    WITH
      uu = LRN.NormSqr(u)
    DO
      IF uu = 0.0d0 THEN
        FOR i := 0 TO LAST(g) DO
          xDt[i] := 0.0d0;
          uDt[i] := - g[i]
        END
      ELSE
        WITH b = LRN.Dot(g, u)/(gNorm*uu) DO
          FOR i := 0 TO LAST(g) DO
            xDt[i] := - b*u[i];
            uDt[i] := - g[i] - u[i];
          END
        END
      END
    END
  END Diff4;

(*********************************************************************)

TYPE 
  Path5 = Path OBJECT OVERRIDES 
      eqs := Eqs5; size := Size5; start := Start5; diff := Diff5
    END;

PROCEDURE Eqs5(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN 
      "dx/dt = u" & "; " &
      "du/dt = - g - |g| u"
  END Eqs5;
  
PROCEDURE Size5(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN 2*NC END Size5;
  
PROCEDURE Start5(<*UNUSED*> p: Path;
    READONLY g: Gradient; gNorm: Slope; VAR u: LRN.T
  ): Speed =
  BEGIN 
    u := g; 
    LRN.Scale(-1.0d0/gNorm, u);
    RETURN 1.0d0
  END Start5;
  
PROCEDURE Diff5(<*UNUSED*> p: Path; 
    READONLY g: Gradient; gNorm: Slope;
    READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN
    xDt := u;
    FOR i := 0 TO LAST(g) DO
      uDt[i] := - g[i] - gNorm * u[i]
    END
  END Diff5;
  
(*********************************************************************)

TYPE 
  Path6 = Path OBJECT OVERRIDES 
      eqs := Eqs6; size := Size6; start := Start6; diff := Diff6
    END;

PROCEDURE Eqs6(<*UNUSED*> p: Path): TEXT =
  BEGIN 
    RETURN 
      "dx/dt = - g - (|g|/|u|)u" & "; " &
      "du/dt = - g - u"
  END Eqs6;
  
PROCEDURE Size6(<*UNUSED*> p: Path; NC: CARDINAL): CARDINAL =
  BEGIN RETURN 2*NC END Size6;
  
PROCEDURE Start6(<*UNUSED*> p: Path;
    READONLY g: Gradient; gNorm: Slope; VAR u: LRN.T
  ): Speed =
  BEGIN
    u := g;
    RETURN gNorm
  END Start6;
  
PROCEDURE Diff6(<*UNUSED*> p: Path; 
    READONLY g: Gradient; gNorm: Slope;
    READONLY u: LRN.T; VAR xDt, uDt: LRN.T
  ) =
  BEGIN
    WITH
      uNorm = LRN.Norm(u)
    DO
      xDt := g;
      LRN.Neg(xDt);
      uDt := xDt;
      IF uNorm > 0.0d0 THEN
        WITH b = gNorm/uNorm DO
          FOR i := 0 TO LAST(g) DO
            xDt[i] := xDt[i] - b*u[i];
            uDt[i] := uDt[i] - u[i]
          END
        END
      END
    END
  END Diff6;
  
BEGIN
  path[1] := NEW(Path1);
  path[2] := NEW(Path2);
  path[3] := NEW(Path3);
  path[4] := NEW(Path4);
  path[5] := NEW(Path5);
  path[6] := NEW(Path6);
END GradMinimizer.
