MODULE JSUniMin;

FROM UniMin IMPORT EvalProc, CheckProc, ComputeDerivatives;
IMPORT Wr, Thread, Fmt;
FROM Stdio IMPORT stderr;

REVEAL 
  T = Public BRANDED OBJECT 
      alpha: LONGREAL := 0.3819660112d0;
    OVERRIDES
      setAlpha := SetAlpha;
      name := Name;
      needsDerivative := NeedsDerivative;
      minimize := Minimize
    END;
    
PROCEDURE SetAlpha(jum: T; alpha: LONGREAL): T =
  BEGIN
    jum.alpha := alpha;
    RETURN jum
  END SetAlpha;
  
PROCEDURE NeedsDerivative(<*UNUSED*> jum: T): BOOLEAN =
  BEGIN
    RETURN FALSE   
  END NeedsDerivative;

PROCEDURE Name(jum: T): TEXT =
  BEGIN
    RETURN "JSUniMin(alpha = " & Fmt.LongReal(jum.alpha) & ")"
  END  Name;

PROCEDURE Minimize(
    um: T;
    eval: EvalProc;
    VAR x: LONGREAL;
    VAR fx: LONGREAL;
    VAR dfx: LONGREAL;
    tol: LONGREAL;
    dist: LONGREAL;
    a: LONGREAL;
    b: LONGREAL;
    check: CheckProc;
  ) =
  VAR 
    u, fu: LONGREAL;    (* Lower bracketing point *)
    m, fm: LONGREAL;    (* Middle point *)
    v, fv: LONGREAL;    (* Upper bracketing point *)
    
  PROCEDURE PrintStatus () =
    BEGIN 
      PrintNewLine();
      PrintAbscissa("a", a);
      PrintPoint("u", u, fu);
      PrintPoint("m", m, fm);
      PrintPoint("v", v, fv);
      PrintAbscissa("b", b);
      PrintNewLine();
    END PrintStatus;
  
  CONST 
    Eps = 1.0d-6;  (* Square root of relative machine precision. *)

  VAR
    s, fs: LONGREAL;      (* Splitting point. *)
    dp, ddp, q: LONGREAL; (* Estimated derivatives at "x" are "dp/q" and "ddp/q". *)
    dfs: LONGREAL;        (* Derivative of "f" at "s" (returned but ignored) *)
    xOld: LONGREAL;       (* Previous value of "x" *)
  BEGIN
    WITH
      jum = NARROW(um, T),
      debug = jum.debug,
      Alpha = jum.alpha,
      Beta = (1.0d0 - Alpha)/Alpha,
      Ateb = Alpha/(1.0d0-Alpha),
      htol = tol/2.0d0,
      qtol = tol/5.0d0
    DO
      <* ASSERT a < x *>
      <* ASSERT b > x *>
      <* ASSERT tol > 0.0d0 *>
      <* ASSERT Alpha < 0.5d0 *>
      
      m := x; fm := fx; 

      dp := 0.0d0; ddp := 0.0d0; q := 0.0d0;
      xOld := x;
      
      u := MAX(a, m - dist - tol - Eps*ABS(m)); 
      eval(u, fu, dfs);
      IF fu < fx THEN x := u; fx := fu; dfx := dfs END;

      v := MIN(b, m + dist + tol + Eps*ABS(m)); 
      eval(v, fv, dfs);
      IF fv < fx THEN x := v; fx := fv; dfx := dfs END;
      
      LOOP
        IF fu >= fm THEN a := u END;
        IF fv >= fm THEN b := v END;

        (* Loop invariants:
            1. Probe points are definitely ordered:
              "u + qtol <= m <= v - qtol"
            2. We are inside the given interval:
              "u >= a, v <= b"
            2. Gaps aren't too unequal:
              "MIN(m-u, v-m) >= Alpha * (v-u)"
        *)

        IF debug THEN PrintStatus() END;

        (* Estimate derivatives at "x" *)
        IF x = u THEN 
          ComputeDerivatives (v, fv, m, fm, x, fx, dp, ddp, q)
        ELSIF x = v THEN                            
          ComputeDerivatives (u, fu, m, fm, x, fx, dp, ddp, q)
        ELSE                                        
          ComputeDerivatives (u, fu, v, fv, x, fx, dp, ddp, q)
        END;
        <* ASSERT q # 0.0d0 *>

        (* Report current points and abort if good enough: *)
        IF x # xOld THEN
          IF check # NIL THEN
            IF check(a, b, x, fx, dp, ddp, q) THEN RETURN END
          END;
          xOld := x;
        END;
        
        (* Check for termination criterion: *)
        IF m - htol <= a AND m + htol >= b THEN RETURN END;

        (* Estimate minimum position: *)
          
        IF ddp <= 0.0d0 THEN
          IF debug THEN Message("convex triple") END;
          IF dp < 0.0d0 THEN
            s := LAST(LONGREAL)
          ELSIF dp > 0.0d0 THEN
            s := FIRST(LONGREAL)
          ELSE
            s := x
          END
        ELSE
          s := x - dp/ddp
        END;

        WITH
          auHi = u - MAX(qtol, Beta*(m-u)),
          umLo = u + MAX(qtol, Alpha*(m-u)),
          umHi = m - MAX(qtol, MAX(Ateb*(v-m), Alpha*(m-u))),
          mvLo = m + MAX(qtol, MAX(Ateb*(m-u), Alpha*(v-m))),
          mvHi = v - MAX(qtol, Alpha*(v-m)),
          vbLo = v + MAX(qtol, Beta*(v-m))
        DO
          IF umLo > umHi AND mvLo > mvHi THEN
            Message("no space to split!");
            PrintStatus();
            PrintAbscissa("umLo = ", umLo);
            PrintAbscissa("umHi = ", umHi);
            PrintAbscissa("mvLo = ", mvLo);
            PrintAbscissa("mvHi = ", mvHi);
          END;
          <* ASSERT (umLo <= umHi) OR (mvLo <= mvHi) *>
          (* Adjust estimate to preserve loop invariants: *)
          IF s < u THEN
            (* Try to probe before "u" *)
            IF a <= auHi THEN
              s := MAX(a, MIN(auHi, s))
            ELSE
              IF debug THEN Message("can't split [a_u], splitting [u_v]") END;
              IF umLo <= umHi THEN s := umLo ELSE s := mvLo END
            END
          ELSIF s > v THEN
            (* Try to probe after "v" *)
            IF vbLo <= b THEN
              s := MIN(b, MAX(vbLo, s))
            ELSE
              IF debug THEN Message("can't split [v_b], splitting [u_v]") END;
              IF mvLo <= mvHi THEN s := mvHi ELSE s := umHi END
            END
          ELSIF s <= m THEN
            IF umLo <= umHi THEN
              s := MAX(umLo, MIN(umHi, s))
            ELSE
              IF debug THEN Message("can't split [u_m], splitting [m_v]") END;
              s := mvLo
            END
          ELSE
            IF mvLo <= mvHi THEN
              s := MAX(mvLo, MIN(mvHi, s))
            ELSE
              IF debug THEN Message("can't split [m_v], splitting [u_m]") END;
              s := umHi
            END
          END;
        END;

        (* Evaluate new point: *)
        eval(s, fs, dfs);
        IF fs < fx THEN
          x := s; fx := fs; dfx := dfs
        END;

        (* Discard one point and rename others: *)
        IF s < u THEN
          IF debug THEN Message("extrapolating down") END;
          v := m; fv := fm;
          m := u; fm := fu;
          u := s; fu := fs;
        ELSIF s > v THEN
          IF debug THEN Message("extrapolating up") END;
          u := m; fu := fm;
          m := v; fm := fv;
          v := s; fv := fs;
        ELSIF fs < fm THEN
          IF s <= m THEN 
            IF debug THEN Message("new minimum in [u__m]") END;
            v := m; fv := fm;
          ELSE 
            IF debug THEN Message("new minimum in [m__v]") END;
            u := m; fu := fm;
          END;
          m := s; fm := fs;
        ELSE
          IF s <= m THEN 
            IF debug THEN Message("high point in [u__m]") END;
            u := s; fu := fs;
          ELSE 
            IF debug THEN Message("high point in [m__v]") END;
            v := s; fv := fs;
          END
        END
      END
    END
  END Minimize;

PROCEDURE Message (msg: TEXT) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, "\n");
  END Message;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

PROCEDURE PrintAbscissa (msg: TEXT; u: LONGREAL) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN 
    Wr.PutText(stderr, "    ");
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(u), 12));
    Wr.PutText(stderr, "\n");
  END PrintAbscissa;

PROCEDURE PrintPoint (msg: TEXT; u, fu: LONGREAL; dfu, ddfu: LONGREAL := 0.0d0) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN 
    Wr.PutText(stderr, "    ");
    Wr.PutText(stderr, msg);
    Wr.PutText(stderr, " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(u), 12));
    Wr.PutText(stderr, " f" & msg & " = ");
    Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(fu), 12));
    IF dfu # 0.0d0 OR ddfu # 0.0d0 THEN
      Wr.PutText(stderr, " df" & msg & " = ");
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(dfu), 18));
      Wr.PutText(stderr, " ddf" & msg & " = ");
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(ddfu), 18));
    END;
    Wr.PutText(stderr, "\n");
  END PrintPoint;

BEGIN
END JSUniMin.
