MODULE RKF4Integrator;

IMPORT Math, Integrator;
FROM Math IMPORT sqrt;
FROM Integrator IMPORT DiffProc, Time, Coord, State, Velocity, Error;

REVEAL T = Public BRANDED OBJECT
    v, v2, v3, v4, v5, v6: REF Velocity;
    sb: REF State;
    er: REF Error;
  OVERRIDES
    name := Name;
    setSize := SetSize;
    computeStep := ComputeStep;
    integrate := Integrate;
  END;

PROCEDURE Name(<*UNUSED*> ri: T): TEXT =
  BEGIN
    RETURN "RKF4Integrator"
  END  Name;

PROCEDURE SetSize(ri: T; n: CARDINAL): Integrator.T =
  BEGIN
    ri.v := NEW(REF Velocity, n);
    ri.sb := NEW(REF State, n);
    ri.er := NEW(REF Error, n);
    ri.v2 := NEW(REF Velocity, n);
    ri.v3 := NEW(REF Velocity, n);
    ri.v4 := NEW(REF Velocity, n);
    ri.v5 := NEW(REF Velocity, n);
    ri.v6 := NEW(REF Velocity, n);
    RETURN ri
  END SetSize;

PROCEDURE Integrate(
    int: T;
    t: Time;
    VAR s: State;
    diff: DiffProc;
    dt: Time;
    checkStep: Integrator.CheckStepProc;
    checkState: Integrator.CheckStateProc := NIL;
  ): Time =
  VAR tb, tf: Time;
  BEGIN
    WITH
      v = int.v^,
      v2 = int.v2^,
      v3 = int.v3^,
      v4 = int.v4^,
      v5 = int.v5^,
      v6 = int.v6^,
      sb = int.sb^,
      er = int.er^
    DO
      tf := t + dt;
      WHILE t < tf DO
        diff(t, s, v);
        IF checkState # NIL THEN
          tf := MIN(tf, checkState(t, s, v));
          IF tf <= t THEN RETURN t END;
        END;
        (* Tries to advance until some time "t" in "(t __ tf]": *)
        REPEAT
          tb := tf;
          Step(diff, t, s, v, tb, sb, er, v2, v3, v4, v5, v6);
          tf := checkStep(t, s, v, tb, sb, er);
          IF tf <= t THEN RETURN t END
        UNTIL tf >= tb;
        t := tb; 
        s := sb;
      END;
    END;
    RETURN t
  END Integrate;

PROCEDURE ComputeStep(<*UNUSED*> int: T; dt: Time; error, tol: Coord): Time =
  BEGIN
    IF error > 0.0d0 THEN
      WITH delta = sqrt(sqrt(tol/error)) DO
        IF delta <= 0.1d0  THEN dt := 0.1d0 * dt
        ELSIF delta >= 10.0d0  THEN dt := 10.0d0 * dt
        ELSE dt := delta * dt
        END
      END
    ELSE
      dt := 10.0d0 * dt;
    END;
    RETURN dt
  END ComputeStep;

PROCEDURE Step(
    diff: DiffProc;
    ta: Time;
    READONLY sa: State;
    READONLY va: Velocity;
    tb: Time;
    VAR sb: State; (* Final state *)
    VAR er: Error; (* Error estimate *)
    VAR v2, v3, v4, v5, v6: Velocity; (* Auxiliary velocities: *)
  ) =
  CONST
    C12: LONGREAL = 1.0d0/4.0d0;
    Ct2: LONGREAL = C12;
    C13: LONGREAL = 3.0d0/32.0d0;
    C23: LONGREAL = 9.0d0/32.0d0;
    Ct3: LONGREAL = C13 + C23;
    C14: LONGREAL = 1932.0d0/2197.0d0;
    C24: LONGREAL = -7200.0d0/2197.0d0;
    C34: LONGREAL = 7296.0d0 /2197.0d0 ;
    Ct4: LONGREAL = C14 + C24 + C34;
    C15: LONGREAL = 439.0d0 /216.0d0 ;
    C25: LONGREAL = -8.0d0 ;
    C35: LONGREAL = 3680.0d0 /513.0d0 ;
    C45: LONGREAL = -845.0d0 /4104.0d0 ;
    Ct5: LONGREAL = C15 + C25 + C35 + C45;
    C16: LONGREAL = -8.0d0 /27.0d0 ;
    C26: LONGREAL = 2.0d0 ;
    C36: LONGREAL = -3544.0d0 /2565.0d0 ;
    C46: LONGREAL = 1859.0d0 /4104.0d0 ;
    C56: LONGREAL = -11.0d0 /40.0d0 ;
    Ct6: LONGREAL = C16 + C26 + C36 + C46 + C56;
    Ce1: LONGREAL = 1.0d0 /360.0d0 ;
    Ce3: LONGREAL = -128.0d0 /4275.0d0 ;
    Ce4: LONGREAL = -2197.0d0 /75240.0d0 ;
    Ce5: LONGREAL = 1.0d0 /50.0d0 ;
    Ce6: LONGREAL = 2.0d0 /55.0d0 ;
    Cv1: LONGREAL = 25.0d0 /216.0d0 ;
    Cv3: LONGREAL = 1408.0d0 /2565.0d0 ;
    Cv4: LONGREAL = 2197.0d0 /4104.0d0 ;
    Cv5: LONGREAL = -1.0d0 /5.0d0 ;
  BEGIN
    WITH
      n = NUMBER(sa),
      dt = tb - ta
    DO
      (* Degree 1 extrapolation: *)
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] + dt * C12*va[i];
      END;

      (* Degree 2 extrapolation: *)
      diff(ta + Ct2 * dt, sb, v2);
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] + dt * (C13*va[i] + C23*v2[i]);
      END;

      (* Degree 3 extrapolation: *)
      diff(ta + Ct3 * dt, sb, v3);
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] + dt * (C14*va[i] + C24*v2[i] + C34*v3[i]);
      END;

      (* Degree 4 extrapolation: *)
      diff(ta + Ct4 * dt, sb, v4);
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] + dt * (C15*va[i] + C25*v2[i] + C35*v3[i] + C45*v4[i]);
      END;

      (* Degree 5 extrapolation: *)
      diff(ta + Ct5 * dt, sb, v5);
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] + 
          dt * (C16*va[i] + C26*v2[i] + C36*v3[i] + C46*v4[i] + C56*v5[i]);
      END;

      (* Error estimation: *)
      diff(ta + Ct6 * dt, sb, v6);
      FOR i := 0 TO n-1 DO
        er[i] := Ce1*va[i] + Ce3*v3[i] + Ce4*v4[i] + Ce5*v5[i] + Ce6*v6[i]
      END;

      (* Compute final state: *)
      FOR i := 0 TO n-1 DO
        sb[i] := sa[i] +
          dt * (Cv1*va[i] + Cv3*v3[i] + Cv4*v4[i] + Cv5*v5[i]);
      END

    END
  END Step;

BEGIN 
END RKF4Integrator.
 
