MODULE WaltzMinimizer;

IMPORT LRN, Minimizer; 
FROM Minimizer IMPORT 
  Point, Vector, Gradient, Length, Value, Slope, EvalProc, TraceProc;

IMPORT Wr, Fmt, Thread, Math;
FROM Stdio IMPORT stderr;

REVEAL
  T = Public BRANDED OBJECT
      gamma: LONGREAL := 1.0d0;
      setSizeCalled: BOOLEAN := FALSE;
      NC: CARDINAL;     (* Number of client variables *)
      xn: REF Point;    (* (Work) probe point *)
      gn: REF Gradient; (* (Work) its gradient *)
      xo: REF Point;    (* First accepted point of current pass *)
      dx: REF Vector;   (* Displacement of current pass *)
    OVERRIDES
      name := Name;
      setGamma := SetGamma;
      setSize := SetSize;
      minimize := Minimize;
      needsGradient := NeedsGradient;
    END;
    
PROCEDURE Name(m: T): TEXT =
  BEGIN
    RETURN "WaltzMinimizer("
     & "gamma = " & Fmt.LongReal(m.gamma)
     & ")"
  END  Name;

PROCEDURE SetGamma(m: T; gamma: LONGREAL): T =
  BEGIN
    <* ASSERT gamma >= 0.0d0 *>
    m.gamma := gamma;
    RETURN m
  END SetGamma;
    
PROCEDURE SetSize(m: T; n: CARDINAL): Minimizer.T =
  BEGIN
    m.setSizeCalled := TRUE;
    m.NC := n;
    m.xn := NEW(REF Point, m.NC);
    m.gn := NEW(REF Gradient, m.NC);
    m.xo := NEW(REF Point, m.NC);
    m.dx := NEW(REF Vector, m.NC);
    RETURN m
  END SetSize;

PROCEDURE NeedsGradient(<*UNUSED*> m: T): BOOLEAN =
  BEGIN
    RETURN TRUE   
  END NeedsGradient;

PROCEDURE Minimize(
    m: T;
    eval: EvalProc;
    VAR x: Point;
    VAR f: Value;
    VAR g: Gradient;
    dist: Length;
    flat: Slope;
    tol: Length;
    maxEvals: CARDINAL;
    trace: TraceProc;
  ) =
  VAR gNorm: Slope;                 (* Gradient norm at current best point *)
      fn: Value;                    (* Function value at probe point *)
      gnNorm: Slope;                (* Gradient norm at probe point *)
      nEvals: CARDINAL := 0;        (* Counts calls to the "eval" procedure. *)
      phase: CARDINAL := 0;         (* Counts acceptances since last diagonal step *)
      step: LONGREAL := dist;       (* Current stepsize *)
      diagonal: BOOLEAN;            (* Last probe was a diagonal one *)
  BEGIN
    <* ASSERT m.setSizeCalled *>
    IF maxEvals <= 0 THEN RETURN END;
    WITH 
      NC = m.NC,
      debug = m.debug,
      alpha = 3.06956450765297882d0,   (* Max step expansion *)
      dzeta = 1.0d0/alpha,             (* Min step contraction *)
      beta =  0.5d0,                   (* Max step contraction *)
      apb = (alpha + beta)/2.0d0,
      amb = (alpha - beta)/2.0d0,
      bpz = (beta + dzeta)/2.0d0,
      bmz = (beta - dzeta)/2.0d0,
      gamma = m.gamma,
      xn = m.xn^,
      gn = m.gn^,
      xo = m.xo^,
      dx = m.dx^,
      minStep = tol/dzeta,   (* Min reducible step *)
      maxStep = dist/alpha   (* Max expandible step *)
    DO
      <* ASSERT NUMBER(x) = NC *>
      <* ASSERT NUMBER(g) = NC *>
      
      gNorm := LRN.Norm(g);
      IF debug THEN 
        Debug("initial state:", LRN.T{});
        Debug("  f, gNorm  =   ", LRN.T{f, gNorm});
        Debug("  x  = ", x); 
        Debug("  g  = ", g);
        PrintNewLine();
      END;

      WHILE nEvals < maxEvals AND gNorm > flat DO

        (* "x", "f", and "g" contain the best minimum seen so far. *)

        IF debug THEN 
          Debug("  nEvals = " & Fmt.Int(nEvals) & " phase = " & Fmt.Int(phase), LRN.T{});
        END;
            
        (* Compute new probe point: *)
        IF gamma = 0.0d0 OR phase < NC+2 THEN
          (* Perform a plain downhill step *)
          IF debug THEN 
            Debug("  ** plain: step = ", LRN.T{step}); 
          END;
          WITH 
            s = step/gNorm
          DO
            FOR i := 0 TO NC-1 DO xn[i] := x[i] - s * g[i] END
          END;
          diagonal := FALSE
        ELSE
          (* Try just one diagonal step *)
          FOR i := 0 TO NC-1 DO
            dx[i] := x[i] - xo[i];
            xn[i] := x[i] + gamma * dx[i]
          END;
          IF debug THEN Debug("  ** diagonal: step = ", LRN.T{LRN.Norm(dx)}) END;
          phase := 0;
          diagonal := TRUE
        END;
        eval(xn, fn, gn); INC(nEvals);
        gnNorm := LRN.Norm(gn);
        IF trace # NIL THEN trace(xn, fn, gn) END;

        IF debug THEN 
          Debug("  fn, gnNorm = ", LRN.T{fn, gnNorm}); 
          Debug("  xn = ", xn); 
          Debug("  gn = ", gn);
        END;

        (* Decide what to do: *)
        IF fn < f THEN
          (* Accept and adjust stepsize: *)
          IF debug THEN Debug("  accepted", LRN.T{}) END;
          IF NOT diagonal AND gnNorm > flat AND step < maxStep THEN 
            WITH 
              cos = MAX(-1.0d0, MIN(+1.0d0, LRN.Dot(g, gn)/(gNorm * gnNorm))),
              aux = (1.0d0 + cos)*(1.0d0 + cos)/4.0d0
            DO
              step := step * (dzeta + (alpha - dzeta)*aux)
            END
          END;
          x := xn; f := fn; g := gn; gNorm := gnNorm;
          INC(phase); 
          IF phase = 1 THEN xo := x END;
        ELSIF NOT diagonal AND step <= minStep THEN
          (* Give up: *)
          IF debug THEN Debug("  rejected and stopped", LRN.T{}) END;
          RETURN
        ELSE
          (* Reduce step size and retry: *)
          IF debug THEN Debug("  rejected", LRN.T{}) END;
          IF NOT diagonal AND step > minStep THEN 
            IF gnNorm > flat THEN
              WITH 
                cos = MAX(-1.0d0, MIN(+1.0d0, LRN.Dot(g, gn)/(gNorm * gnNorm)))
              DO
                step := step * (bpz - bmz*cos)
              END
            ELSE
              step := 0.5d0 * step
            END;
          END;
        END;
        IF debug THEN PrintNewLine() END;
      END
    END
  END  Minimize;
  
PROCEDURE Debug(msg: TEXT; READONLY x: LRN.T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "WaltzMinimizer ");
    Wr.PutText(stderr, msg);
    FOR i := 0 TO LAST(x) DO
      Wr.PutText(stderr, Fmt.Pad(Fmt.LongReal(x[i], Fmt.Style.Sci, 6), 15))
    END;
    Wr.PutText(stderr, "\n")
  END Debug;

PROCEDURE PrintNewLine () =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    Wr.PutText(stderr, "\n");
  END PrintNewLine;

BEGIN
END WaltzMinimizer.



